//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  olf.cpp
//! \brief COlo, COld, COld online files

#include "defs.hpp"

#include <fstream>
#include <sstream>
#include <exception>
#include <algorithm>
#include <memory>
#include <fcntl.h>
#include <sys/stat.h>
#include <boost/format.hpp>

#include "../trivia/file_ops.hpp"
#include "../readplus/ask.hpp"

#include "olf.hpp"
#include "slice.hpp"
#include "genus.hpp"
#include "obj.hpp"
#include "expr.hpp"
#include "xax_lex.hpp"

using boost::format;


//**************************************************************************************************
//*  class COlo
//**************************************************************************************************


//! Remove one z.

void COlo::remove_z( int iz )
{
    ZCo.erase( ZCo.begin()+iz );
    for( int j=0; j<nJ(); ++j )
        V[j]->z.erase( V[j]->z.begin()+iz );
}


//! Line containing z values and curve parameters.

string COlo::info_line( int j ) const
{
    const int LEN=15;
    char wrd[LEN];
    snprintf( wrd, LEN, "%2i%c ", j, V[j]->frozen ? '/' : ' ' );
    string out = wrd;
    out += V[j]->info_line();
    return out;
}


//! Return value of iz-th z-variable for slice j.

RObj COlo::z( int j, int iz ) const
{
    return V[j]->z[iz];
}


//**************************************************************************************************
//*  trivially duplicated COld/COlc functions
//**************************************************************************************************


//! Duplicate spectrum j, and return new shared pointer.

PSlice COld::new_slice( int j ) const
{
    return PSlice( new CSpec( *(VS(j)) ) );
}


//! Duplicate curve j, and return new shared pointer.

PSlice COlc::new_slice( int j ) const
{
    return PSlice( new CCurve( *(VC(j)) ) );
}


//! New shared pointer to copied online file (but without copying main vector).

POlo COld::new_olo( bool modified ) const
{
    POlo ret( new COld( *this ) );
    ret->V.clear();
    ret->as_on_disk = as_on_disk && !modified;
    return ret;
}

POlo COlc::new_olo( bool modified ) const
{
    POlo ret( new COlc( *this ) );
    ret->V.clear();
    ret->as_on_disk = as_on_disk && !modified;
    return ret;
}

POld COld::new_old( bool modified ) const
{
    return P2D( new_olo( modified ) );
}

POlc COlc::new_olc( bool modified ) const
{
    return P2C( new_olo( modified ) );
}


//! Set main vector by copying from fin.

void COld::copy_mainvec( ROlo fin )
{
    ROld fd = R2D( fin );
    V.resize( fin->nJ() );
    for( int j=0; j<fin->nJ(); ++j )
        V[j] = PSlice( new CSpec( *(fd->VS(j)) ) );
}


//! Set main vector by copying from fin.

void COlc::copy_mainvec( ROlo fin )
{
    ROlc fc = R2C( fin );
    V.resize( fin->nJ() );
    for( int j=0; j<fin->nJ(); ++j )
        V[j] = PSlice( new CCurve( *(fc->VC(j)) ) );
}


//! Return pointer to selected main vector entry.

PSpec COld::VS( int j ) const
{
    if( j<0 || j>=V.size() )
        throw "VS call with j=" + S(j) + " though V.size=" + S(V.size());
    PSpec s = std::dynamic_pointer_cast<class CSpec>( V[j] );
    if ( !s )
        throw "BUG: file "+S(j)+" does not contain data";
    return s;
}

PCurve COlc::VC( int j ) const
{
    if( j<0 || j>=V.size() )
        throw "VC call with j=" + S(j) + " though V.size=" + S(V.size());
    PCurve c = std::dynamic_pointer_cast<class CCurve>( V[j] );
    if ( !c )
        throw "BUG: file "+S(j)+" does not contain curves";
    return c;
}


//! Return one coordinate.

CCoord COld::coord( const CGenus& genus ) const
{
    if        ( genus.categ==VCateg::X ) {
        return xco;
    } else if ( genus.categ==VCateg::Y ) {
        return yco;
    } else if ( genus.categ==VCateg::Z ) {
        if( genus.num >= nZ() )
            throw "coordinate " + genus.genus_info() + " not defined";
        return ZCo[genus.num];
    } else if ( genus.categ==VCateg::R ) {
        if( genus.num >= RPar.size() )
            throw "coordinate " + genus.genus_info() + " not defined";
        return RPar[genus.num].Co;
    } else {
        throw "variable " + genus.genus_info() + " not defined for curve file";
    }
}


//! Return one coordinate.

CCoord COlc::coord( const CGenus& genus ) const
{
    if        ( genus.categ==VCateg::X ) {
        return xco;
    } else if ( genus.categ==VCateg::Y ) {
        return yco;
    } else if ( genus.categ==VCateg::Z ) {
        if( genus.num >= nZ() )
            throw "coordinate " + genus.genus_info() + " not defined";
        return ZCo[genus.num];
    } else if ( genus.categ==VCateg::CP ) {
        if( genus.num >= nP )
            throw "coordinate " + genus.genus_info() + " not defined";
        return PCo[genus.num];
    } else if ( genus.categ==VCateg::R ) {
        if( genus.num >= RPar.size() )
            throw "coordinate " + genus.genus_info() + " not defined";
        return RPar[genus.num].Co;
    } else {
        throw "variable " + genus.genus_info() + " not defined for curve file";
    }
}


//! Change one coordinate.

void COld::set_coord( const CGenus& genus, CCoord& co )
{
    if        ( genus.categ==VCateg::X ) {
        xco = co;
    } else if ( genus.categ==VCateg::Y ) {
        yco = co;
    } else if ( genus.categ==VCateg::Z ) {
        if( genus.num >= nZ() )
            throw "coordinate " + genus.genus_info() + " not defined";
        ZCo[genus.num] = co;
    } else if ( genus.categ==VCateg::R ) {
        if( genus.num >= RPar.size() )
            throw "coordinate " + genus.genus_info() + " not defined";
        RPar[genus.num].Co = co;
    } else {
        throw "variable " + genus.genus_info() + " not defined for curve file";
    }
}


//! Change one coordinate.

void COlc::set_coord( const CGenus& genus, CCoord& co )
{
    if        ( genus.categ==VCateg::X ) {
        xco = co;
    } else if ( genus.categ==VCateg::Y ) {
        yco = co;
    } else if ( genus.categ==VCateg::Z ) {
        if( genus.num >= nZ() )
            throw "coordinate " + genus.genus_info() + " not defined";
        ZCo[genus.num] = co;
    } else if ( genus.categ==VCateg::CP ) {
        if( genus.num >= nP )
            throw "coordinate " + genus.genus_info() + " not defined";
        PCo[genus.num] = co;
    } else if ( genus.categ==VCateg::R ) {
        if( genus.num >= RPar.size() )
            throw "coordinate " + genus.genus_info() + " not defined";
        RPar[genus.num] = co;
    } else {
        throw "variable " + genus.genus_info() + " not defined for curve file";
    }
}


//**************************************************************************************************
//*  class COld
//**************************************************************************************************


//! Create data file based on curve file.

COld::COld( class COlc const* c )
{
    name = c->name;
    lDoc = c->lDoc;
    xco = c->xco;
    yco = c->yco;
    ZCo = c->ZCo;
    RPar = c->RPar;
}


//! Remove unused error bars.

void COld::purge_dy()
{
    for ( int j=0; j<nJ(); ++j ) {
        if ( !VS(j)->has_dy() )
            return; // dy already absent, nothing to do
        for ( int i=0; i<VS(j)->size(); ++i) {
            if( VS(j)->dy[i] != 0 )
                return; // dy has at least one nonzero entry, nothing to do
        }
    }
    for ( int j=0; j<nJ(); ++j )
        VS( j )->dy.clear();
}


//! Return number of points in spectrum j.

int COld::nPts( int j ) const
{
    if ( !nJ() || j>=nJ() )
        throw "BUG: nPts called with invalid j="+S(j);
    return VS(j)->size();
}


//! Return number of points if it is the same for all spectra. Else return 0.

int COld::nPts() const
{
    if ( !nJ() )
        throw "BUG: nPts() called while nSpec=0";
    int np = VS(0)->size();
    for ( int j=0; j<nJ(); ++j )
        if (VS(j)->size()!=np) return 0;
    return np;
}


//! Returns true if all data points have nonzero dy.

bool COld::has_nonzero_dy() const
{
    return std::all_of( V.cbegin(), V.cend(),
                        [](RSlice s){ return (PCAST<const CSpec>(s))->has_nonzero_dy(); } );
}

//**************************************************************************************************
//*  class COlc - interactive initialization
//**************************************************************************************************

string COlc::wgtNames[] = { "lin", "log", "var", "varc", "vard" };

//! Converts weight from string to enum

COlc::TWgt COlc::name2wgt( string name )
{
    for ( int i=0; i<=_LAST_TWGT; ++i )
        if( name==COlc::wgtNames[i] )
            return (TWgt)i;
    throw "undefined weighing key '" + name + "'";
}

//! Prompt for curve description, and initialize things.

void COlc::curve_query( const string& quest )
{ // EMBEDDED_DIALOG
 ask_again:
    string resp = sask( quest+" (or e(c|i)(r|w); h for help)" );
    if        ( resp=="h" ) {
        cout <<
            "curve definition:\n"
            "- either an arithmetic expression,\n"
            "    containing the running genusiable t,\n"
            "    for instance: p0+p1*t\n"
            "- or external evaluation mode e(c|i)(r|w)\n"
            "       next question will ask for external program\n"
            "    c: program reads pars from command line\n"
            "    i: program reads #pars, pars from standard input\n"
            "    r: program reads #points, points from standard input\n"
            "    w: program writes default grid, Frida interpolates\n";
        goto ask_again;
    } else if ( resp.size()==3 && resp[0]=='e' &&
                ( resp[1]=='c' || resp[1]=='i' ) &&
                ( resp[2]=='r' || resp[2]=='w' ) ) {
        if      ( resp[1]=='c' )
            scrInpMode = COlc::_CMDLIN;
        else if ( resp[1]=='i' )
            scrInpMode = COlc::_STDIN;
        else
            throw "BUG: impossible case";
        if      ( resp[2]=='r' )
            evaMode = COlc::_SCR;
        else if ( resp[2]=='w' )
            evaMode = COlc::_SCR_INTP;
        else
            throw "BUG: impossible case";
        expr = sask( "Curve evaluation command" );
        nP = iask( "Number of parameters" );
    } else {
        parse_function( resp );
    }
    if( nP>100 )
        throw "PLAUSIBILITY CHECK FAILED: unexpected number of parameters np = " + S(nP);
}


//**************************************************************************************************
//*  class COlc - low-level initialization
//**************************************************************************************************


//! Sets auxiliary parameters to default values.

void COlc::curve_set_defaults()
{
    for( int ip=0; ip<nP; ip++ )
        PCo.push_back( CCoord("p" + S(ip), "") );
    weighing = COlc::_VAR;
}


//! Clears chi^2 and similar fit outcomes, to be called when parameters are manually changed.

void COlc::clear_fits_outcome()
{
    chi2 = 0;
    for( int j=0; j<nJ(); ++j )
        VC(j)->clear_fit_outcome();
}


//! Parses function expression.

void COlc::parse_function( const string& _expr )
{
    evaMode = COlc::_EXPR;
    expr = _expr;
    T = user_xaxparse( expr.c_str() );
    if( !T->has_dummy() )
        cout << "WARNING: curve has no formal argument t\n";
    nP = T->npar();
}


//**************************************************************************************************
//*  class COlc - function evaluation
//**************************************************************************************************


//! Vectorial evaluation of a curve given by an expression tree.

RObjVec COlc::eval_curve_expr( const vector<double>& vt, int k, int j, bool want_error ) const
{
    CContext ctx( k, j );
    ctx.want_error = want_error;
    RObj ret;
    if        ( !T->has_dummy() ) {  // curve does not depend on t
        ret = T->tree_val( ctx );
    } else {
        ctx.request_VT( &vt );
        ret = T->tree_val( ctx );
    }
    if ( !ret->is_vec() )
        throw S("BUG in eval_curve_expr: T->tree_val didn't return a vector");
    return PCAST<const CObjVec>(ret);
}


//! Vectorial evaluation of a curve given by a system call.

RObjVec COlc::eval_curve_syscall( const vector<double>& vt, int k, int j,
                                  bool want_error /* currently ignored */ ) const
{
    // Create FIFO to receive return values.
          // TODO: choose name as: triv::next_tmp_file( "/tmp/frida.%i" );
    string ret_fifonam = string("/tmp/ret-") + getenv( "LOGNAME" );
    triv::system( "rm -f "+ret_fifonam );
    if ( mkfifo( ret_fifonam.c_str(), 0666 ) )
        throw "SYSTEM ERROR cannot make " + ret_fifonam;

    string cmd = expr;
    string inp_fifo = "";
    string errmsg;
    string par_fifonam;

    // Create FIFO for passing pars or/and grid via stdin.
    if ( scrInpMode==_STDIN || evaMode == _SCR ) {
        par_fifonam = string("/tmp/par-") + getenv( "LOGNAME" );
        triv::system( "rm -f "+par_fifonam );
        if ( mkfifo( par_fifonam.c_str(), 0666 ) )
            throw "SYSTEM ERROR cannot make " + par_fifonam;
        inp_fifo = " < " + par_fifonam;
    }
    // Parameters passed on command line.
    if ( scrInpMode==_CMDLIN ){
        string pars = "";
        for ( int ip=0; ip<nP; ++ip )
            pars += " " + str( format("%g ") % VC(j)->P[ip] );
        cmd += pars;
    }

    // Start the external process.
    cmd += inp_fifo + " > " + ret_fifonam + " &";
    errmsg = "command '" + cmd + "'";
    system( cmd.c_str() );

    // Feed parameters and x values to FIFO.
    if( inp_fifo!="" ){
        int par_fifo;
        // we use open instead of fopen or ofstream,
        // because we need non-blocking mode.
        if ( !( par_fifo = open( par_fifonam.c_str(), O_WRONLY ) ) )
            throw "SYSTEM ERROR cannot open " + par_fifonam;
        fcntl( par_fifo, F_SETFL, O_NONBLOCK);

        if ( scrInpMode==_STDIN ){
            string out = str( format("%i ") % nP );
            write( par_fifo, out.c_str(), out.size() );
            for ( int ip=0; ip<nP; ++ip ) {
                string out = str( format("%16.8g ") % VC(j)->P[ip] );
                write( par_fifo, out.c_str(), out.size() );
            }
        }
        if ( evaMode==_SCR ) {
            string out = str( format("%i ") % vt.size() );
            write( par_fifo, out.c_str(), out.size() );
            for ( int i=0; i<vt.size(); ++i ) {
                string out = str( format("%16.8g ") % vt[i] );
                write( par_fifo, out.c_str(), out.size() );
            }
        }
        close( par_fifo );
    }
    //cout << "DEB cvv: #vt=" << vt.size() << ", called " << errmsg << "\n";

    std::ifstream FF( ret_fifonam );
    if( !FF.is_open() )
        throw "BUG: cannot read results of system call from " + ret_fifonam;

    vector<double> v( vt.size() );

    if        ( evaMode==_SCR ) {
        for ( int i=0; i<vt.size(); ++i ) {
            string line;
            if( !getline( FF, line ) )
                throw "BUG: failed to read line " + S(i) + " from " + ret_fifonam;
            std::istringstream iss(line);
            double x;
            try {
                iss >> x >> v[i];
            } catch( std::exception &ex ) {
                throw S(ex.what()) + ": returns no x y pair for t["+S(i)+"]="+
                    S(vt[i])+" but '"+line+"'";
            }
            if( fabs(x-vt[i])>1e-4*fabs(vt[i]+x) )
                throw errmsg + " returns x=" + S(x) + " instead of " +
                    S(vt[i]) + " in line '"+line+"'";
            if ( !std::isfinite(v[i]) )
                throw "infinite y value in line '"+line+"'";
        }
    } else if ( evaMode==_SCR_INTP ) {
        CSpec S;
        double x, y;
        string line;
        while( getline( FF, line ) ) {
            std::istringstream iss(line);
            try {
                iss >> x >> y;
            } catch( std::exception &ex ) {
                throw S(ex.what()) + ": invalid line '"+line+"'";
            }
            S.push_xy( x, y );
        }
        // interpolation:
        if ( S.size()<2 )
            throw errmsg + " returns only " + S(S.size()) + " valid lines";
        //cout << "DEB cvv: intpol "<<vt.front()<<" .. "<<vt.back()<<"\n";
        S.intpol( vt, &v, nullptr );
    } else
        throw S("BUG: invalid evaMode");
    FF.close();
    return PObjVecDbl( new CObjVecDbl( v ) );
}


//! Vectorial evaluation of a curve.

RObjVec COlc::eval_curve( const vector<double>& vt, int k, int j, bool want_error ) const
{
    if ( evaMode==_EXPR )
        return eval_curve_expr( vt, k, j, want_error );
    else if ( evaMode==_SCR || evaMode==_SCR_INTP )
        return eval_curve_syscall( vt, k, j, want_error );
    else
        throw S("BUG: unexpected evaluation mode");
}


//! Scalar evaluation of a curve.

RObj COlc::eval_curve_scalar( double arg, int k, int j, bool want_error ) const
{
    vector<double> vt( 1, arg );
    RObjVec cu = eval_curve( vt, k, j, want_error );
    if ( cu->size()!=1 )
        throw "BUG: vector has "+S(cu->size())+" elements while exactly one was expected";
    return cu->to_obj(0);
}


//**************************************************************************************************
//*  class COlc - inspect
//**************************************************************************************************

//! Returns text lines to describe file-wide settings.

vector<string> COlc::info_settings() const
{
    vector<string> ret;
    string out = "";
    if      ( evaMode==_SCR )
        out = "script: ";
    else if ( evaMode==_SCR_INTP )
        out = "script (followed by interpolation): ";
    ret.push_back( out + expr );
    string txt = "";
    if( kd!=-1 )
        txt += "data file: " + S(kd) + ", ";
    if( kconv!=-1 )
        txt += "conv file: " + S(kconv) + ", ";
    txt += "weighing: " + weight_str();
    ret.push_back( txt );
    if( chi2>0 )
        ret.push_back( "global fit result: "+S(chi2));
    return ret;
}


//! Returns header line for info_line's.

string COlc::info_table_header() const
{
    string ret = "j ";
    for ( int iz=0; iz<nZ(); iz++ )
        ret += str( format( "  z%i          " ) % iz );
    for ( int ip=0; ip<nP; ip++ )
        ret += str( format( "  p%i           " ) % ip );
    ret += " oc chi^2       1-R^2";
    return ret;
}


//! Prints a table of j,z,p lines.

void COlc::print_info_table() const
{
    for( string line: info_settings() )
        cout << line << "\n";
    cout << info_table_header() << "\n";
    for( int j=0; j<nJ(); ++j )
        cout << info_line( j ) << "\n";
}


//! Returns true if some parameters are global.

bool COlc::has_global_par() const
{
    for ( int j=0; j<nJ(); ++j )
        if ( VC(j)->has_global_par() )
            return true;
    return false;
}


//! Returns description of fit weight.

string COlc::weight_str() const
{
    switch ( weighing ) {
    case _LIN:
        return "constant";
    case _LOG:
        return "logarithmic";
    case _VAR:
        return "with reciprocal genusiance (data and curve)";
    case _VARC:
        return "with reciprocal genusiance (curve only)";
    case _VARD:
        return "with reciprocal genusiance (data only)";
    }
    throw S("BUG: unexpected weight");
}
