//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  opr.cpp
//! \brief NOperate: functional operations.

#include "defs.hpp"

#include <boost/format.hpp>

#include "../readplus/ask.hpp"
#include "../readplus/macro.hpp"

#include "olf.hpp"
#include "mem.hpp"
#include "slice.hpp"
#include "variables.hpp"
#include "genus.hpp"
#include "obj.hpp"
#include "expr.hpp"
#include "opr.hpp"
#include "xax_lex.hpp"

using boost::format;


//**************************************************************************************************
//*  operate via generic expressions
//**************************************************************************************************


//! Display data. Use boolean expression to display only a subset of the data.

void NOperate::show( const string& subcmd )
{
    NOlm::IterateD fiter;

    static string expr = "1";
    expr = sask( "Show if", expr);
    if (expr=="")
        return;
    RNode T = user_xaxparse( expr.c_str() );
    if( T->has_dummy() )
        throw S("dummy argument t not allowed when operating on data");

    while( ROld fin=fiter() ) {
        if ( fiter.size()>1 )
            printf( ".. file %i:\n", fiter.k() );

        string outj, outp;

        // determine length of longest spectrum:
        int ni = 0;
        for ( int j=0; j<fin->nJ(); j++)
            ni = std::max( ni, fin->VS(j)->size() );

        // printout by i:
        for ( int i=0; i<ni; i++) {
            outj = "";
            outp = "";

            for( int j=0; j<fin->nJ(); j++ ){
                RSpec s = fin->VS(j);
                if( i>= s->size() ) continue;
                if ( ! T->tree_point_dbl( fiter.k(), j, i ) )
                    continue;
                outj += S(j) + " ";
                if      ( subcmd=="xyd" )
                    if( s->has_dy() )
                        outp += str( format( "%- 12.5g %- 12.5g +- %- 12.5g" )
                                     % s->x[i] % s->y[i] % s->dy[i] );
                    else
                        outp += str( format( "%- 12.5g %- 12.5g" ) % s->x[i] % s->y[i] );
                else if ( subcmd=="y" )
                    outp += str( format( "%- 12.5g" ) % s->y[i] );
                else
                    throw "invalid subcmd " + subcmd;
                outp += " ";
            }

            if( outp=="" )
                continue;

            printf( "%5i [%s] %s\n", i, outj.substr(0,outj.size()-1).c_str(),
                    outp.substr(0,outp.size()-1).c_str() );
        }
    }
}


//! Delete or retain data points according to a boolean expression.

void NOperate::select( bool sel_del )
{
    NOlm::IterateO fiter;
    static string expr;
    expr = sask((sel_del ? "Delete if" : "Retain if"), expr);
    if (expr=="")
        return;
    RNode T = user_xaxparse( expr.c_str() );
    if( T->has_dummy() )
        throw S("dummy argument t not allowed when operating on data");

    while ( ROlo fin=fiter() ) {
        ROld fd = R2D( fin );
        ROlc fc = R2C( fin );

        POlo fout( fin->new_olo() );
        fout->lDoc.push_back((sel_del ? "md " : "mr ") + expr);

        for ( int j=0; j<fin->nJ(); j++ ) {
            if ( fd ) {
                RSpec ein = fd->VS(j);
                PSpec eout = PSpec( new CSpec );
                eout->z = ein->z;
                vector<double> lvec( ein->size() );
                T->tree_vec_val( &lvec, 0, fiter.k(), j );
                bool with_dy = ein->has_dy();
                for ( int i=0; i<ein->size(); i++ ) {
                    if ( sel_del ^ (bool) lvec[i] ) {
                        if( with_dy )
                            eout->push_xyd( ein->x[i], ein->y[i], ein->dy[i] );
                        else
                            eout->push_xy( ein->x[i], ein->y[i] );
                    }
                }
                if ( eout->size() )
                    fout->V.push_back( eout );
            } else if ( fc ) {
                PCurve eout = PCurve( new CCurve( *(fc->VC(j)) ) );
                if ( sel_del ^ T->tree_point_val( fiter.k(), j )->to_b() )
                    fout->V.push_back( eout );
            }
        }
        if( !fout->nJ() )
            throw S("no data point left");
        NOlm::mem_store( fout, fiter.k() );
    }
}


//! Pointwise operation on x, y, or z<n>, according to arithmetic expression.

void NOperate::Pointwise( string llabel )
{
    NOlm::IterateO fiter;

    CGenus lref(llabel);

    static string expr;
    expr = sask("Expression", expr);
    if (expr=="")
        return;
    RNode T = user_xaxparse( expr.c_str() );
    if( T->has_dummy() )
        throw S("dummy argument t not allowed when operating on data");

    while( ROlo fin=fiter() ){
        int k = fiter.k();
        ROld fd = R2D( fin );
        ROlc fc = R2C( fin );

        if ( lref.pointwise() && !fd )
            throw S("no pointwise operation on curve");
        if( lref.categ==VCateg::CP && !fc )
            continue;

        POlo fout( fin->new_olo() );

        if( lref.categ!=VCateg::CP )
            fout->lDoc.push_back( "o" + lref.genus_info() + " " + expr);

        CCoord co = T->node_coord( k );
        if        ( lref.categ==VCateg::X ) {
            fout->xco = co;
        } else if ( lref.categ==VCateg::Y ) {
            fout->yco = co;
        } else if ( lref.categ==VCateg::Z ) {
            if( lref.num==-1 )
                fout->ZCo.push_back( co );
            else if( lref.num >= fin->ZCo.size() )
                throw S("no such z coordinate to operate on");
            else
                fout->ZCo[lref.num] = co;
        } else if ( lref.categ==VCateg::R ) {
            if( lref.num==-1 )
                fout->RPar.push_back( CParam( co ) );
            else if( lref.num >= fin->RPar.size() )
                throw S("no such real par to operate on");
            else
                fout->RPar[lref.num].Co = co;
        } else if ( lref.categ==VCateg::CP ) {
            if( !fc )
                continue; // just ignore op operations on data files
            if( lref.num>=fc->nP )
                throw S("no such parameter to operate on");
            NOlm::overwrite = 1;
            P2C(fout)->clear_fits_outcome();
        }

        if ( lref.categ==VCateg::R ) {
            fout->copy_mainvec( fin );
            double val = T->tree_point_val( k, 0 )->to_r();
            if( lref.num==-1 )
                fout->RPar.back().val = val;
            else
                fout->RPar[lref.num].val = val;
        } else {
            fout->V.resize( fin->nJ() );
            for( int j=0; j<fin->nJ(); j++ ){
                fout->V[j] = fin->new_slice( j );
                if ( fin->V[j]->frozen )
                    continue;
                if ( lref.pointwise() ) {
                    PSpec eout( new CSpec( *(fd->VS(j)) ) );
                    vector<double> vout( fd->nPts(j) );
                    if ( lref.categ==VCateg::X || lref.categ==VCateg::DY ) {
                        T->tree_vec_val( &vout, 0, k, j );
                        if      ( lref.categ==VCateg::X )
                            (P2D(fout))->VS(j)->x = vout;
                        else if ( lref.categ==VCateg::DY )
                            (P2D(fout))->VS(j)->dy = vout;
                        else
                            throw S("BUG: unexpected lref");
                    } else if ( lref.categ==VCateg::Y ) {
                        vector<double> dvout( fd->nPts(j) );
                        T->tree_vec_val( &vout, &dvout, k, j );
                        (P2D(fout))->VS(j)->y = vout;
                        if( dvout.size() )
                            (P2D(fout))->VS(j)->dy = dvout;
                        else
                            (P2D(fout))->VS(j)->dy.clear();
                    } else
                        throw S("BUG: unexpected lref");
                } else if ( lref.categ==VCateg::Z ) {
                    RObj pret = T->tree_point_val( k, j );
                    if ( lref.num==-1 )
                        fout->V[j]->z.push_back( pret );
                    else
                        fout->V[j]->z[lref.num] = pret;
                } else if ( lref.categ==VCateg::CP ) {
                    (P2C(fout))->VC(j)->P[lref.num] = T->tree_point_val( k, j )->to_r();
                } else {
                    throw S("BUG: invalid oxy");
                }
            }
        }
        NOlm::mem_store( fout, fiter.k() );
    }
}


//! Integral operation, using arithmetic expression.

void NOperate::Integral()
{
    NOlm::IterateO fiter;

    static string expr;
    expr = sask( "Functional", expr );
    if ( expr=="" )
        return;
    RNode T = user_xaxparse( expr.c_str() );
    if( T->has_dummy() )
        throw S("dummy argument t not allowed when operating on data");

    while( ROlo fin=fiter() ) {
        int k = fiter.k();
        int nz = fin->ZCo.size();
        bool savable = nz > 0;

        POld fout;
        PSpec sout;
        if ( savable ) {
            if      ( ROlc fc = R2C(fin) )
                fout = POld( new COld( fc.get() ) );
            else
                fout = POld( (R2D(fin))->new_old() );
            fout->lDoc.push_back( "oi " +expr );
            fout->ZCo.pop_back();
            fout->xco = fin->ZCo.back();
            fout->yco = T->node_coord( k );
            sout = PSpec( new CSpec );
            sout->z = fin->V[0]->z;
            sout->z.pop_back();
        }

        for ( int j=0; j<fin->nJ(); j++ ) {
            if ( savable && fin->V[j]->frozen )
                continue;
            double xval;
            RObj res = T->tree_point_val( k, j );
            if (!savable) {
                cout << "result: " << res->to_s() << "\n";
                continue;
            }
            xval = fin->V[j]->z[nz-1]->to_r();
            sout->push_xyd( xval, res->to_r(), res->to_dr() );
            if (nz>=2) { // new spectrum if jump in other z values
                double zval = fin->V[j]->z[nz-2]->to_r();
                if ( j+1 < fin->nJ() && fin->V[j+1]->z[nz-2]->to_r() != zval ) {
                    fout->V.push_back( sout );
                    sout = PSpec( new CSpec );
                    sout->z = fin->V[j+1]->z;
                    sout->z.pop_back();
                }
            }
        }
        if ( savable ) {
            fout->purge_dy();
            fout->V.push_back( sout );
            NOlm::mem_store( fout, fiter.k() );
        }
//      if ( !savable )
//          NReg::lastresult = res;
    }
}


//! Integral operations to generate not only new y, but also new x.

void NOperate::IntXY( string mode )
{ // EMBEDDED_DIALOG
    NOlm::IterateD fiter;

    bool with_d;
    if      ( mode=="xy" )
        with_d = false;
    else if ( mode=="xyd" )
        with_d = true;
    else
        throw S("invalid command");

    static int icolx = 0, icoly = 1, icold = 2;
    static CCoord xco, yco;
    icolx = iask( "x from column", icolx );
    if( icolx<0) return;
    icoly = iask( "y from column", icoly );
    if( icoly<0) return;
    if ( with_d ) {
        icold = iask( "dy from column", icold );
        if( icold<0) return;
    }
    xco = CCoord( "x", "" );
    yco = CCoord( "y", "" );

    while( ROld fin=fiter() ) {
        int nz = fin->ZCo.size();

        POld fout( fin->new_old() );
        string lin = S("oi") + mode + " " + S(icolx) + " " + S(icoly);
        if ( with_d )
            lin += " " + S(icold);
        fout->lDoc.push_back( lin );

        fout->xco = xco;
        fout->yco = yco;

        PSpec sout( new CSpec );

        for ( int j=0; j<fin->nJ(); j++ ) {
            int n = fin->nPts(j);
            if( icolx>=n || icoly>=n || (with_d && icold>=n) )
                throw "not enough columns in spectrum " + S(j);
            if ( with_d )
                sout->push_xyd( fin->VS(j)->y[icolx],
                                fin->VS(j)->y[icoly],
                                fin->VS(j)->y[icold] );
            else
                sout->push_xy( fin->VS(j)->y[icolx],
                               fin->VS(j)->y[icoly] );
            sout->z = fin->VS(j)->z;
            if ( nz ) { // new spectrum if jump in any z values
                if ( j+1 < fin->nJ() && fin->V[j+1]->z != fin->V[j]->z ) {
                    fout->V.push_back( sout );
                    sout = PSpec( new CSpec );
                }
            }
        }
        fout->V.push_back( sout );
        NOlm::mem_store( fout, fiter.k() );
    }
}


//**************************************************************************************************
//*  functional operations
//**************************************************************************************************

//! Functional integration, derivation, ...

void NOperate::Functional( const string& subcmd )
{
    NOlm::IterateD fiter;
    while ( ROld fin=fiter() ) {
        POld fout( fin->new_old() );
        fout->lDoc.push_back( "of"+subcmd );

        if        ( subcmd=="i" ) { // integrate
            fout->yco.name = "Int d" + fin->xco.name + " " + fin->yco.name;
            fout->yco.unit = fin->xco.unit + " * " + fin->yco.unit;
        } else if ( subcmd=="s" ) { // sum
            fout->yco.name = "Sum_0^i " + fin->yco.name;
            fout->yco.unit = fin->yco.unit;
        } else if ( subcmd=="d" ) { // derive
            fout->yco.name = "d" + fin->yco.name + " / d" + fin->xco.name;
            fout->yco.unit = fin->yco.unit + " / " + fin->xco.unit;
        } else if ( subcmd=="ac" ) { // autocorrelate
            fout->yco.name = "ac{" + fin->yco.name + "}";
            fout->yco.unit = fin->yco.unit + "^2";
        } else {
            throw S("invalid 'of' command");
        }

        for ( int j=0; j<fin->nJ(); j++ ) {
            PSpec sin = fin->VS(j);
            PSpec sout( new CSpec );
            sout->z = sin->z;
            int n = sin->size();

            if        ( subcmd=="i" ) { // integrate
                sout->x = sin->x;
                sout->y.resize( n );
                sout->y[0] = 0;
                for ( int i=1; i<n; ++i )
                    sout->y[i] = sout->y[i-1] +
                        (sin->x[i] - sin->x[i-1]) *
                        (sin->y[i] + sin->y[i-1]) / 2;
            } else if ( subcmd=="s" ) { // sum
                sout->x = sin->x;
                sout->y.resize( n );
                sout->y[0] = sin->y[0];
                for ( int i=1; i<n; ++i )
                    sout->y[i] = sout->y[i-1] + sin->y[i];
            } else if ( subcmd=="d" ) { // derive
                for( int i=0; i<n-1; ++i ){
                    double dx = sin->x[i+1] - sin->x[i];
                    sout->push_xy(
                        (sin->x[i+1] + sin->x[i] ) / 2,
                        dx==0 ? 0 : (sin->y[i+1] - sin->y[i]) / dx );
                }
            } else if ( subcmd=="ac" ) { // autocorrelate
                throw "currently disabled, old code in old/src-fragm/opr-func-ac.cpp";
            }
            fout->V.push_back( sout );
        }
        NOlm::mem_store( fout, fiter.k() );
    }
}
